/* DWARF2 EH unwinding support for x86 Windows.
   Copyright (C) 2006 Free Software Foundation, Inc.

This file is part of GCC.

GCC is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

In addition to the permissions in the GNU General Public License, the
Free Software Foundation gives you unlimited permission to link the
compiled version of this file with other programs, and to distribute
those programs without any restriction coming from the use of this
file.  (The General Public License restrictions do apply in other
respects; for example, they cover modification of the file, and
distribution when not linked into another program.)

GCC is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GCC; see the file COPYING.  If not, write to
the Free Software Foundation, 51 Franklin Street, Fifth Floor,
Boston, MA 02110-1301, USA.  */

/* Do code reading to identify a signal frame, and set the frame
   state data appropriately.  See unwind-dw2.c for the structs.

   The basic idea is to detect that we are indeed trying to unwind past a
   signal handler and to fill out the GCC internal unwinding structures for
   the OS kernel frame as if it had been directly called from the interrupted
   context.

   This is all assuming that the code to set the handler asked the kernel to
   pass a pointer to such context information.

   There is three main parts.

   1) The first thing to do is to check if we are in a signal context. If not
      we can just return as there is nothing to do.  We are probably on some
      foreign code for which no unwind frame can be found.  If this is a call
      from the Windows signal handler, then:

   2) We must get the signal context information. 

      * With the standard exception filter:

      This is on Windows pointed to by an EXCEPTION_POINTERS.  We know that the
      signal handle will call an UnhandledExceptionFilter with this
      parameter.  The spec for this routine is:

         LONG WINAPI UnhandledExceptionFilter(struct _EXCEPTION_POINTERS*);

      So the pointer to struct _EXCEPTION_POINTERS must be somewhere on the
      stack.

      This was found experimentally to always be at offset 0 of the context
      frame in all cases handled by this implementation.

      * With the SEH exception handler:

      In this case the signal context is directly on the stack as the SEH
      exception handler has the following prototype:

         DWORD
         SEH_error_handler (PEXCEPTION_RECORD ExceptionRecord,
                            PVOID EstablisherFrame,
                            PCONTEXT ContextRecord,
                            PVOID DispatcherContext)

      This was found experimentally to always be at offset 56 of the context
      frame in all cases handled by this implementation.

   3) When we have the signal context we just have to save some registers and
      set the return address based on the program counter (Eip).

   Note that this implementation follows closely the same principles as the
   GNU/Linux and OSF ones.  */


/* Patterns found experimentally to be in a Windows signal handler.  */

/* In a standard exception filter.  */

#define SIG_PAT1 \
  (pc[-2] == 0xff && pc[-1] == 0xd0     /* call %eax           */ \
   && pc[0] == 0x83 && pc[1] == 0xf8)   /* cmp 0xdepl,%eax     */

#define SIG_PAT2 \
  (pc[-5] == 0xe8 && pc[-4] == 0x68   /* call (depl16)       */ \
   && pc[0] == 0xc3)                  /* ret                 */

/* In a Win32 SEH handler.  */

#define SIG_SEH1 \
  (pc[-5] == 0xe8                     /* call addr           */ \
   && pc[0] == 0x83 && pc[1] == 0xc4  /* add 0xval,%esp      */ \
   && pc[3] == 0xb8)                  /* mov 0xval,%eax      */

#define SIG_SEH2 \
  (pc[-5] == 0x8b && pc[-4] == 0x4d   /* mov depl(%ebp),%ecx */ \
   && pc[0] == 0x64 && pc[1] == 0x8b) /* mov %fs:(0),<reg>   */ \

/* In the GCC alloca (stack probing).  */

#define SIG_ALLOCA \
  (pc[-1] == 0x83                    /* orl $0x0,(%ecx)     */ \
   && pc[0] == 0x9 && pc[1] == 0                               \
   && pc[2] == 0x2d && pc[3] == 0    /* subl $0x1000,%eax   */ \
   && pc[4] == 0x10 && pc[5] == 0)

#define MD_FALLBACK_FRAME_STATE_FOR x86_fallback_frame_state

/* GCC hard register numbers and mapping to DW2 column numbers.  We
   unfortunately can't reuse the back-end definitions because they
   might refer to symbols (e.g. register map tables) not linked into
   libgcc.  */

#define GCC_EAX 0
#define GCC_EBX 3
#define GCC_ECX 2
#define GCC_EDX 1
#define GCC_ESI 4
#define GCC_EDI 5
#define GCC_EBP 6
#define GCC_ESP 7

/* Partial .eh_frame mapping for the GCC registers above, in accordance with
   DWARF_FRAME_REGNUM+DWARF2_FRAME_REG_OUT definitions.  */

static int const libgcc_dbx_register_map[] =
{
  /* Indexed by GCC REGNO ...
                0   1   2   3   4   5   6   7    */
  /* GCC */ /*  EAX EDX ECX EBX ESI EDI EBP ESP  */
  /* DW  */     0,  2,  1,  3,  6,  7,  4,  5
};

#define DWFR(REGNO) libgcc_dbx_register_map[REGNO]

#define DW_EIP 8  /* retaddr column */

static _Unwind_Reason_Code
x86_fallback_frame_state (struct _Unwind_Context *context,
			  _Unwind_FrameState *fs)
{
  char *eh_debug_env = getenv ("EH_DEBUG");
  int eh_debug = eh_debug_env ? atoi (eh_debug_env) : 0;
 
  void *ctx_ra  = (void *)(context->ra);  /* return address */
  void *ctx_cfa = (void *)(context->cfa); /* context frame address */
  unsigned char *pc = (unsigned char *) ctx_ra;

  if (eh_debug)
    printf ("FALLBACK called for CFA = 0x%p, RA = 0x%p\n", ctx_cfa, ctx_ra);

  /* In the test below we look for two specific patterns found
     experimentally to be in the Windows signal handler.  */
  if (SIG_PAT1 || SIG_PAT2 || SIG_SEH1 || SIG_SEH2)
    {
      PEXCEPTION_POINTERS weinfo;
      PCONTEXT proc_ctx;
      long new_cfa;

      if (SIG_SEH1) 
	proc_ctx = (PCONTEXT) (*(int*)(ctx_cfa + 56));
      else if (SIG_SEH2)
	proc_ctx = (PCONTEXT) (*(int*)(ctx_cfa + 8));
      else
	{
	  weinfo = (PEXCEPTION_POINTERS) (*(int*)ctx_cfa);
	  proc_ctx = weinfo->ContextRecord;
	}

      /* The new context frame address is the stack pointer.  */
      new_cfa = proc_ctx->Esp;
      fs->cfa_how = CFA_REG_OFFSET;
      fs->cfa_reg = __builtin_dwarf_sp_column();
      fs->cfa_offset = new_cfa- (long) ctx_cfa;

      /* Save some registers.  */
      fs->regs.reg[DWFR(GCC_EAX)].how = REG_SAVED_OFFSET;
      fs->regs.reg[DWFR(GCC_EAX)].loc.offset = (long)&proc_ctx->Eax - new_cfa;
      fs->regs.reg[DWFR(GCC_EBX)].how = REG_SAVED_OFFSET;
      fs->regs.reg[DWFR(GCC_EBX)].loc.offset = (long)&proc_ctx->Ebx - new_cfa;
      fs->regs.reg[DWFR(GCC_ECX)].how = REG_SAVED_OFFSET;
      fs->regs.reg[DWFR(GCC_ECX)].loc.offset = (long)&proc_ctx->Ecx - new_cfa;
      fs->regs.reg[DWFR(GCC_EDX)].how = REG_SAVED_OFFSET;
      fs->regs.reg[DWFR(GCC_EDX)].loc.offset = (long)&proc_ctx->Edx - new_cfa;
      fs->regs.reg[DWFR(GCC_ESI)].how = REG_SAVED_OFFSET;
      fs->regs.reg[DWFR(GCC_ESI)].loc.offset = (long)&proc_ctx->Esi - new_cfa;
      fs->regs.reg[DWFR(GCC_EDI)].how = REG_SAVED_OFFSET;
      fs->regs.reg[DWFR(GCC_EDI)].loc.offset = (long)&proc_ctx->Edi - new_cfa;
      fs->regs.reg[DWFR(GCC_EBP)].how = REG_SAVED_OFFSET;
      fs->regs.reg[DWFR(GCC_EBP)].loc.offset = (long)&proc_ctx->Ebp - new_cfa;

      fs->regs.reg[9].how = REG_SAVED_OFFSET;
      fs->regs.reg[9].loc.offset = (long)&proc_ctx->Eip - new_cfa;

      /* Set the return address to Eip + 1.  As we can be called multiple
	 times we use another register for this.  */
      proc_ctx->Dr0 = proc_ctx->Eip + 1;
      fs->regs.reg[DW_EIP].how = REG_SAVED_OFFSET;
      fs->regs.reg[DW_EIP].loc.offset = (long)&proc_ctx->Dr0 - new_cfa;
      fs->retaddr_column = DW_EIP;

      return _URC_NO_REASON;
    }

  /* Unwinding through _alloca, propagating from a trap triggered by
     one of it's probes prior to the real SP adjustment.  The only
     operations of interest performed is "pushl %ecx", followed by
     ecx clobbering.  */
  else if (SIG_ALLOCA) 
    {
      /* Only one push between entry in _alloca and the probe trap.  */ 
      long new_cfa = (long) ctx_cfa + 4;

      fs->cfa_how = CFA_REG_OFFSET;
      fs->cfa_reg = __builtin_dwarf_sp_column();
      fs->cfa_offset = new_cfa - (long) ctx_cfa;

      /* The saved value of %ecx is at CFA - 4 */
      fs->regs.reg[DWFR(GCC_ECX)].how = REG_SAVED_OFFSET;
      fs->regs.reg[DWFR(GCC_ECX)].loc.offset = -4;

      /* And what is stored at the CFA is the return address.  */
      fs->regs.reg[DW_EIP].how = REG_SAVED_OFFSET;
      fs->regs.reg[DW_EIP].loc.offset = 0;
      fs->retaddr_column = DW_EIP;

      return _URC_NO_REASON;
    }

  else
    return _URC_END_OF_STACK;  
}
